######################################################################
package HarnChar;
######################################################################

=head1 NAME

HarnChar -- object for reading, holding, and storing a RuneQuest character

=head1 SYNOPSIS

    use HarnChar;
    $char= new HarnChar();
    $char= HarnChar->rollAll();
    $char->readFile($file);
    $char->writeFile(); # writes to $char->shortname().".txt"
    $char->writeFile($file);
    @statnames = $char->stats();
    $value = $char->$stat;          # return a stat value
    $char->$stat($newvalue);        # set a stat value
    $char->validate($printflag);

=head1 DESCRIPTION

This is an object-oriented module which stores the character 
information in a set of plain fields.  Each field uses only 
word characters, and corresponds to the JavaScript form 
element name, and the key in the parameter file.  

=cut

require 5.003;
use lib ".";
use strict;
use Carp;
use CGI;
use HM3Data;
use FileHandle;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $AUTOLOAD);

require Exporter;

@ISA     = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT  = qw( );
@EXPORT_OK = qw( );
$VERSION = '0.01';

######################################################################
# private variables
my @statdefs= qw(shortname 		word*25
		 users 			string*40 
		 edition 		word*6 
		 CharacterName 		string*40 
		 PlayerName 		string*40 
		 Species 		Human|Sindarin|Khuzdul
		 Sex 			Male|Female
		 BirthDate 		string*15
		 Sunsign 		Ulandus|Aralius|Feniri|Ahnu|Angberelius|Nadai|Hirin|Tarael|Tai|Skorus|Masara|Lado
		 Birthplace 		word*40
		 Culture		Feudal|Tribal|Viking|Imperial|Sindarin|Khuzdul
		 Occupation		string*40
		 SocialClass 		Slave|Serf|Unguilded|Guilded|Noble
		 SiblingRank 		string*20
		 Parentage 		string*20
		 Estrangement 		string*20 
		 Clanhead 		string*20 
		 Height 		int*3
		 Frame 			Scant|Light|Medium|Heavy|Massive
		 Weight 		int*3
		 Comliness 		int*3
		 Complexion 		string*15
		 HairColor 		string*15
		 EyeColor 		string*15
		 Strength 		int*3
		 Stamina 		int*3
		 Dexterity 		int*3
		 Agility 		int*3
		 Eyesight 		int*3
		 Hearing 		int*3
		 Smell 			int*3
		 Voice 			int*3
		 Intelligence		int*3
		 Aura 			int*3
		 Will 			int*3
		 Morality 		int*3
                 Sexuality		string*15
                 Deity			Agrik|Halea|Ilvir|Larani|Morgath|Naveh|Peoni|Sarajin|Save-K'nor|Siem|Tribal|Godless
		 Piety 			int*3
		 );
my %statdef_of_stat = @statdefs;
my @stats = grep { not m/[\*\|]/ } @statdefs;

my @tuzyn_months = qw( Nuzyael Peonu Kelen Nolus Larane Agrazhar 
                       Azura Halane Savor Ilvin Navek Morgat );

my $data = new HM3Data;

######################################################################
=head1 METHODS

The following methods are available:

=over 4

=item new (Constructor)
Create a new instance of this class by writing
    $mychar = HarnChar->new();
=cut
sub new { 
    my $that = shift;  # either class name or template object
    my $self = {};     # $self is a hash reference
    bless($self, $that);
    return $self;
}

=item copy (Constructor)
Create a new instance of this class by writing
    $char2 = $char->copy();
=cut
sub copy { 
    my $self = shift; # either class name or template object
    my $copy= {};     # $self is a hash reference
    foreach (keys %$self) {
	$copy->{$_} = $self->{$_};
    }
    bless($copy, $self);
    return $copy;
}

=item getChar (Constructor)
Read a character object from a file by 
    $mychar = HarnChar->getChar($shortname);
=cut
sub getChar { 
    my $self = new(shift);
    my $shortname = shift;
    # accept a filename as a shortname, converted by removing the ".txt"
    $shortname =~ s/\.txt$//;
    $self->readFile("$shortname.txt");
    return $self;
}

=item readFile
Read attributes and values into a character object from a file by 
    $mychar->readFile($filename);
=cut
sub readFile { 
    my $self= shift;
    my $filename = shift;
    my $fh = new FileHandle;
    unless (open($fh,$filename)) {
	carp "Problem opening file '$filename'"; 
	return $self;
    }
    my @keylist;
    while (<$fh>) {
	if (/(\w+)(\{?)[\s:=]+(.*)/) {
	    my ($key,$startarray,$value)= ($1,$2,$3);
	    if ($startarray) {
		while (<$fh>) { last if (/^\s*\}/); $value .= $_; }
	    }
	    $self->{$key} = $value;
	    push(@keylist,$_);
	}
    }    
    close($fh);
    $self->{"_HarnChar_keylist"} = \@keylist;
    return $self;
}

=item writeFile
Write attributes and values from a character object to a file by 
    $mychar->writeFile($filename);
=cut
sub writeFile { 
    my $self= shift;
    my $filename = shift || ($self->shortname().".txt");
    my $fh = new FileHandle;
    if (not open($fh,">$filename")) {
	carp "File '$filename' was not found or could not be written to.";
	return $self;
    } else {
	foreach my $stat ($self->stats()) {
	    my $value = $self->$stat();
	    if ($value =~ m/\n/) { # multiline entry
		# remove leading and trailing whitespace including newlines
		$value =~ s/^\s*//s;  $value =~ s/\s*$//s;
		print $fh "$stat\{\n$value\n}\n";
	    } else {
		print $fh "$stat : $value\n";
	    }
	}
    } # end of if (not open(...)) { ... } else { ... }
    return $self;
}

sub statdef {
    my ($self,$stat) = @_;
    return $statdef_of_stat{$stat};
}

sub stats {
    my $self = shift;
    croak "self is not a ref" unless (ref($self));
    # private variables start with an underscore
    my @stats = grep { not m|^_| } keys %{$self};
    my @sorted = sort { _statcmp($a,$b) } @stats;
    return @sorted;
}

sub basestats {
    return qw(Species Species Sex BirthDate Sunsign Birthplace 
	      Culture SocialClass SiblingRank Parentage Occupation
	      Estrangement Clanhead Height 
	      Frame Weight Comliness Complexion HairColor EyeColor 
	      Strength Stamina Dexterity Agility 
	      Eyesight Hearing Smell Voice 
	      Intelligence Aura Will Morality Deity Piety);
}

sub statfield {
    my $self = shift(@_);
    my $req = shift(@_);
    my $name = shift(@_);
    my %opts = @_;
    # now see if there is a defined stat for this
    my $def = $self->statdef($name);
    # if Javascript is enabled, should pass the argument
    # -onChange=>"changeStat(this);"
    # to the textfield routine
    my $html;
    if (not defined($def)) {
	my $size = ($opts{'-size'} or 2);
	my $width = int(0.8*$size)+1;
	$html = $req->textfield(-name=>$name,-value=>$self->$name(),
				-size=>$width);
    } elsif ($def =~ m/^int\*(\d+)\[([\+\-]?\d+):([\+\-]?\d+)\.\.([\+\-]?\d+)\]/) {
	my ($size,$default,$min,$max) = ($1,$2,$3,$4);
	my $value = ($self->$name() or $default);
	$value =~ s/^\s*\+//;
	$html = $req->popup_menu(-name=>$name,-values=>[$min..$max],
				 -default=>$value);
    } elsif ($def =~ m/(string|int)\*(\d+)/) {
	my $size = ($opts{'-size'} or $2);
	my $width = int(0.8*$size)+1;
	$html = $req->textfield(-name=>$name,-value=>$self->$name(),-size=>$width);
    } elsif ($def =~ m/(text)\*(\d+)/) {
	my $size = ($opts{'-size'} or $2);
	my $cols = ($opts{'-cols'} or 60);
	my $rows = int($size/(5*$cols));
	$html = $req->textarea(-name=>$name,-value=>$self->$name(),
			       -rows=>$rows,-columns=>$cols);
    } elsif ($def =~ m/\|/) {
	my @options = split(m/\|/,$def);
	my $default = ($self->$name() or $options[1]);
	$html = $req->popup_menu(-name=>$name,-values=>\@options,
				 -default=>$default);
    }
    return $html;
}

sub editfield {
    my $self = shift(@_);
    my $req = shift(@_);
    my $name= shift(@_);
    my $size= (shift(@_) || 2);
    my $html = statfield($name,$size);
    $html .= $req->button({-value=>"+",-onClick=>
			  "incrementStat(this.form.$name); return false;"});
    $html .= $req->button({-value=>"-",-onClick=>
                          "decrementStat(this.form.$name); return false;"});
    return $html;
}

sub skillTypeList { 
    return (sort $data->getSkillTypeList());
}

sub baseSkill { 
    my ($self,$skillname) = @_;
    $skillname =~ s/\s*\(.*?\)//; # remove specialization
    my $sbcode = $data->getSkillBaseCode($skillname);
    my $mod = $data->getSkillSignCode($skillname);
    my $sb = 0;
    foreach my $code (split(/\s+/,$sbcode)) {
	if ($code eq "CML") {
	    $sb += $self->Comliness();
	} elsif ($code eq "STR") {
	    $sb += $self->Strength();
	} elsif ($code eq "STA") {
	    $sb += $self->Stamina();
	} elsif ($code eq "DEX") {
	    $sb += $self->Dexterity();
	} elsif ($code eq "AGL") {
	    $sb += $self->Agility();
	} elsif ($code eq "EYE") {
	    $sb += $self->Eyesight();
	} elsif ($code eq "HRG") {
	    $sb += $self->Hearing();
	} elsif ($code eq "SML") {
	    $sb += $self->Smell();
	} elsif ($code eq "VOI") {
	    $sb += $self->Voice();
	} elsif ($code eq "INT") {
	    $sb += $self->Intelligence();
	} elsif ($code eq "AUR") {
	    $sb += $self->Aura();
	} elsif ($code eq "WIL") {
	    $sb += $self->Will();
	} else {
	    die "Unknown stat '$code' in '$sbcode' in skill '$skillname'";
	}
    }
    $sb = int( ($sb+1) / 3 ); # round to nearest whole number
    my $sign = $self->Sunsign();
    if (($sign eq "Ulandus") and ($mod =~ m/Ula[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Aralius") and ($mod =~ m/Ara[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Feniri") and ($mod =~ m/Fen[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Ahnu") and ($mod =~ m/Ahn[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Angberelius") and ($mod =~ m/Ang[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Nadai") and ($mod =~ m/Nad[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Hirin") and ($mod =~ m/Hir[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Tarael") and ($mod =~ m/Tar[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Tai") and ($mod =~ m/Tai[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Skorus") and ($mod =~ m/Sko[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Masara") and ($mod =~ m/Mas[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    } elsif (($sign eq "Lado") and ($mod =~ m/Lad[\/\w]*\+?(\-?\d)/)) {
	$sb += $1;
    }
    return $sb;
}

sub addSkill {
    my ($self,$skillname,$base) = @_;
    my $fullname = $skillname;
    $skillname =~ s/\s*\(.*?\)//; # remove specialization
    my $type = $data->getSkillType($skillname);
    my $sb = $self->baseSkill($skillname);
    my $ml;
    if ($base =~ m/^SB(\d)$/) {
	$ml = $sb*$1;
    } elsif ($base =~ m/^(\d\d)\+SB(\d)$/) {
	$ml = $1 + $sb*$2;
    } else {
	die "Invalid base for added skill '$skillname'";
    }
    my $n = $self->skillNumber($skillname);
    my $prefix = "Skill_" . $type . "_" . $n;
    $self->{$prefix."_Name"} = $fullname;
    $self->{$prefix."_Type"} = $type;
    $self->{$prefix."_SB"} = $sb;
    $self->{$prefix."_ML"} = $ml;
    return $self;
}

sub skillNumber {
    my ($self,$skillname) = @_;
    my $fullname = $skillname;
    $skillname =~ s/\s*\(.*?\)//; # remove specialization
    my $type = $data->getSkillType($skillname);
    for (my $n = 1; $n < 999; $n++) {
	my $name = $self->{"Skill_".$type."_".$n."_Name"};
	if (($skillname eq $name) or (not defined($name))) {
	    return $n;
	}
    }
} # end of sub skillNumber

sub skillCount {
    my ($self,$skillname) = @_;
    my $n = 0;
    foreach my $key (keys %$self) {
	$n++ if ($key =~ m/^Skill_\w+_\d+_Name/);
    }
    return $n;
} # end of sub skillNumber

sub startSkills {
    my ($self,$skillname,$base) = @_;
    $self->addSkill("Climbing", "SB4");
    $self->addSkill("Condition", "SB5");
    $self->addSkill("Jumping", "SB4");
    $self->addSkill("Stealth", "SB3");
    $self->addSkill("Throwing", "SB4");
    $self->addSkill("Awareness", "SB4");
    $self->addSkill("Intrigue", "SB3");
    $self->addSkill("Oratory", "SB2");
    $self->addSkill("Rhetoric", "SB3");
    $self->addSkill("Singing", "SB3");
    $self->addSkill("Language: Native", "50+SB1");
    $self->addSkill("Initiative", "SB4");
    $self->addSkill("Unarmed", "SB4");
    return $self;
}

sub addOccupationSkills {
    my ($self) = @_;
    my $occname = $self->Occupation();
    foreach my $skillname ($data->getOccupationSkills($occname)) {
	my $code = $data->getOccupationSkillCode($occname,$skillname);
	$self->addSkill($skillname,$code);
    }
    return $self;
}

sub AUTOLOAD {
    my $self = shift;
    my $type = ref($self) or croak "'$self' is not an object";
    my $name = $AUTOLOAD;
    $name =~ s/.*://; # strip fully-qualified portion of name
    croak "Bad method call '$name'" unless ($name =~ m/^\w/);
    if (@_) {
	my $newval = shift;
	if ($self->{"_HarnChar:printIfChange"} and ($newval ne $self->{$name})) {
	    print "Changing $name from $self->{$name} to $newval\n";
	}
	return $self->{$name} = $newval;
    } else {
	return $self->{$name};
    }
}

######################################################################
sub rollAll {
    my $self = shift(@_);     # $self is a hash reference
    $self->rollSpecies() unless ($self->Species());
    $self->rollSex() unless ($self->Sex());
    $self->rollBirthDate();
    $self->Sunsign( $self->getSunsign($self->BirthDate()) );
    $self->rollSocialClass() unless ($self->SocialClass());
    $self->rollSiblingRank();
    $self->rollParentage();
    $self->rollClanhead();
    $self->rollEstrangement();
    $self->rollHeight();
    $self->rollFrame();
    $self->Weight( $self->getWeight() );
    $self->rollComliness();
    $self->rollComplexion();
    $self->rollHairColor();
    $self->rollEyeColor();
    $self->rollStrength();
    $self->rollStamina();
    $self->rollDexterity();
    $self->rollAgility();
    $self->rollEyesight();
    $self->rollHearing();
    $self->rollSmell();
    $self->rollVoice();
    $self->rollIntelligence();
    $self->rollAura();
    $self->rollWill();
    $self->rollMorality();
    $self->rollSexuality();
    $self->rollDeity() unless ($self->Deity());
    $self->rollPiety();
    $self->rollOccupation() unless ($self->Occupation());
    $self->startSkills();
    $self->addOccupationSkills();
    return $self;
}
######################################################################
sub rollSpecies {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($roll <= 97) {
        $self->Species("Human");
        $self->BaseAge(14);
    } elsif ($roll == 98) {
        $self->Species("Sindarin");
        $self->BaseAge(18);
    } elsif ($roll == 99) {
        $self->Species("Khuzdul");
        $self->BaseAge(16);
    } elsif ($roll == 100) {
        $self->Species("Other");
    }
    return $self;
}
######################################################################
sub rollSex {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($self->Species() eq "Human") {
        if ($roll <= 48) { $self->Sex("Male"); }
        else { $self->Sex("Female"); }
    } elsif ($self->Species() eq "Sindarin") {
        if ($roll <= 45) { $self->Sex("Male"); }
        else { $self->Sex("Female"); }
    } elsif ($self->Species() eq "Khuzdul") {
        if ($roll <= 75) { $self->Sex("Male"); }
        else { $self->Sex("Female"); }
    } else {
        die "Unknown species ".$self->Species();
    }
}
######################################################################
sub rollBirthDate {
    my $self = shift(@_);
    my $dt = $tuzyn_months[$self->roll("1d12")]." ".$self->roll("1d30");
    $self->BirthDate($dt);
    return $self;
}
######################################################################
sub getSunsign {
    my $self = shift(@_);
    my $dt = shift(@_) || $self->BirthDate();
    if (($dt =~ m/^Nuzyael ([5-9]|\d\d)$/) or ($dt =~ m/^Peonu [1-3]$/)) {
        $self->Sunsign("Ulandus");
    } elsif (($dt =~ m/^Peonu ([4-9]|\d\d)$/) or ($dt =~ m/^Kelen [1-2]$/)) {
        $self->Sunsign("Aralius");
    } elsif (($dt =~ m/^Kelen ([3-9]|\d\d)$/) or ($dt =~ m/^Nolus [1-3]$/)) {
        $self->Sunsign("Feniri");
    } elsif (($dt =~ m/^Nolus ([4-9]|\d\d)$/) or ($dt =~ m/^Larane [1-4]$/)) {
        $self->Sunsign("Ahnu");
    } elsif (($dt =~ m/^Larane ([5-9]|\d\d)$/) or ($dt =~ m/^Agrazhar [1-6]$/)) {
        $self->Sunsign("Angberelius");
    } elsif (($dt =~ m/^Agrazhar ([7-9]|\d\d)$/) or ($dt =~ m/^Azura [1-5]$/)) {
        $self->Sunsign("Nadai");
    } elsif (($dt =~ m/^Azura ([6-9]|\d\d)$/) or ($dt =~ m/^Halane [1-4]$/)) {
        $self->Sunsign("Hirin");
    } elsif (($dt =~ m/^Halane ([5-9]|\d\d)$/) or ($dt =~ m/^Savor [1-3]$/)) {
        $self->Sunsign("Tarael");
    } elsif (($dt =~ m/^Savor ([4-9]|\d\d)$/) or ($dt =~ m/^Ilvin [1-2]$/)) {
        $self->Sunsign("Tai");
    } elsif (($dt =~ m/^Ilvin ([3-9]|\d\d)$/) or ($dt =~ m/^Navek [1-2]$/)) {
        $self->Sunsign("Skorus");
    } elsif (($dt =~ m/^Navek ([3-9]|\d\d)$/) or ($dt =~ m/^Morgat 1$/)) {
        $self->Sunsign("Masara");
    } elsif (($dt =~ m/^Morgat ([2-9]|\d\d)$/) or ($dt =~ m/^Nuzyael [1-3]$/)) {
        $self->Sunsign("Lado");
    }
}
######################################################################
sub rollSocialClass {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($self->Species() eq 'Sindarin') {
        if ($roll <= 70) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 99) { $self->SocialClass("Guilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } elsif ($self->Species() eq 'Khuzdul') {
        if ($roll <= 15) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 98) { $self->SocialClass("Guilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } elsif ($self->Culture() eq 'Feudal') {
        if ($roll <= 15) { $self->SocialClass("Slave"); }
        elsif ($roll <= 70) { $self->SocialClass("Serf"); }
        elsif ($roll <= 93) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 98) { $self->SocialClass("Guilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } elsif ($self->Culture() eq 'Tribal') {
        if ($roll <= 10) { $self->SocialClass("Slave"); }
        elsif ($roll <= 99) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } elsif ($self->Culture() eq 'Viking') {
        if ($roll <= 15) { $self->SocialClass("Slave"); }
        elsif ($roll <= 80) { $self->SocialClass("Serf"); }
        elsif ($roll <= 99) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } elsif ($self->Culture() eq 'Imperial') {
        if ($roll <= 25) { $self->SocialClass("Slave"); }
        elsif ($roll <= 90) { $self->SocialClass("Unguilded"); }
        elsif ($roll <= 98) { $self->SocialClass("Guilded"); }
        elsif ($roll <= 100) { $self->SocialClass("Noble"); }
    } else {
        die "Unrecognized culture ".$self->Culture();
    }
    return $self;
}
######################################################################
sub rollSiblingRank {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    $roll -= 20 if ($self->Species() =~ m/(Sindarin|Khuzdul)/);
    if ($roll <= 25) {
        $self->SiblingRank("Eldest");
    } elsif ($roll <= 50) {
        $self->SiblingRank("2nd Child");
    } elsif ($roll <= 70) {
        $self->SiblingRank("3rd Child");
    } elsif ($roll <= 85) {
        $self->SiblingRank("4th Child");
    } elsif ($roll <= 95) {
        $self->SiblingRank("5th Child");
    } elsif ($roll <= 100) {
        $self->SiblingRank("6th Child");
    }
    return $self;
}
######################################################################
sub rollParentage {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    my $subroll = $self->roll("1d100");
    if ($roll <= 50) {
	my $status;
	if ($subroll <= 60) {
	    $status = "Both parents alive and living together";
	} elsif ($subroll <= 70) {
	    $status = "Father Absent";
	} elsif ($subroll <= 75) {
	    $status = "Mother Absent";
	} elsif ($subroll <= 80) {
	    $status = "Father Dead.";
	} elsif ($subroll <= 85) {
	    $status = "Father Dead. Mother remarried.";
	} elsif ($subroll <= 90) {
	    $status = "Mother Dead.";
	} elsif ($subroll <= 100) {
	    $status = "Mother Dead. Father remarried.";
	} else {
	    die;
	}
        $self->Parentage("Offspring ($status)");
    } elsif ($roll <= 70) {
        $self->Parentage("Fostered");
    } elsif ($roll <= 75) {
        $self->Parentage("Adopted");
    } elsif ($roll <= 80) {
        $self->Parentage("Bastard (acknowledged)");
    } elsif ($roll <= 90) {
        $self->Parentage("Bastard (unacknowledged)");
    } elsif ($roll <= 100) {
        $self->Parentage("Orphan");
    } else {
        die;
    }
    return $self;
}
######################################################################
sub rollClanhead {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($roll <= 50) {
        $self->Clanhead("Distant Relation");
    } elsif ($roll <= 75) {
        $self->Clanhead("Aunt or Uncle");
    } elsif ($roll <= 85) {
        $self->Clanhead("Cousin");
    } elsif ($roll <= 100) {
        $self->Clanhead("Father or Mother");
    } else {
	die;
    }
    return $self;
}
######################################################################
sub rollEstrangement {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    $roll += 5 if ($self->SiblingRank() =~ m/eldest/i);
    if ($roll <= 10) {
        $self->Estrangement("Outcast");
    } elsif ($roll <= 40) {
        $self->Estrangement("Unpopular");
    } elsif ($roll <= 60) {
        $self->Estrangement("Average");
    } elsif ($roll <= 95) {
        $self->Estrangement("Popular");
    } else {
        $self->Estrangement("Favorite");
    }
    return $self;
}
######################################################################
sub rollHeight {
    my $self = shift(@_);
    my $roll = $self->roll("4d6");
    $roll += 1 if ($self->Culture() eq 'Viking');
    $roll += 2 if ($self->SocialClass() eq 'Noble');
    if (($self->Species() eq 'Human') and ($self->Sex() eq 'Male')) {
	$self->Height(54+$roll);
    } elsif (($self->Species() eq 'Human') and ($self->Sex() eq 'Female')) {
        $self->Height(52+$roll);
    } elsif (($self->Species() eq 'Sindarin') and ($self->Sex() eq 'Male')) {
        $self->Height(51+$roll);
    } elsif (($self->Species() eq 'Sindarin') and ($self->Sex() eq 'Female')) {
        $self->Height(50+$roll);
    } elsif ($self->Species() eq 'Khuzdul') {
        $self->Height(40+$roll);
    } else {
	die;
    }
    return $self;
}
######################################################################
sub rollFrame {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll -= 3 if (($self->Species() eq 'Human') and ($self->Sex() eq 'Female'));
    $roll -= 2 if ($self->Species() eq 'Sindarin');
    $roll += 3 if ($self->Species() eq 'Khuzdul');
    if ($roll <= 5) {
	$self->Frame("Scant");
    } elsif ($roll <= 8) {
        $self->Frame("Light");
    } elsif ($roll <= 12) {
        $self->Frame("Medium");
    } elsif ($roll <= 15) {
        $self->Frame("Heavy");
    } else {
        $self->Frame("Massive");
    }
    return $self;
}
######################################################################
sub getWeight {
    my $self = shift(@_);
    my $ht = $self->Height();
    my $wt;
    if ($ht <= 51) {
	$wt = (2*$ht)-5;
    } elsif ($ht <= 63) {
	$wt = (3*$ht)-56;
    } elsif ($ht <= 69) {
	$wt = (4*$ht)-119;
    } else {
	$wt = (5*$ht)-190;
    }
    if ($self->Frame() eq 'Scant') {
	$wt = int($wt*0.8);
    } elsif ($self->Frame() eq 'Light') {
	$wt = int($wt*0.9);
    } elsif ($self->Frame() eq 'Heavy') {
	$wt = int($wt*1.1);
    } elsif ($self->Frame() eq 'Massive') {
	$wt = int($wt*1.2);
    }
    return $wt;
}
######################################################################
sub rollComliness {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    if ($self->{'verbose-comliness'}) {
	if ($roll <= 5) {
	    $self->Comliness("$roll - Ugly");
	} elsif ($roll <= 8) {
	    $self->Comliness("$roll - Plain");
	} elsif ($roll <= 12) {
	    $self->Comliness("$roll - Average");
	} elsif ($roll <= 15) {
	    $self->Comliness("$roll - Attractive");
	} else {
	    $self->Comliness("$roll - Handsome");
	}
    } else {
	$self->Comliness($roll);
    }
    return $self;
}
######################################################################
sub rollComplexion {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($self->Species() eq 'Human') {
	if ($roll <= 27) {
	    $self->Complexion("Fair");
	} elsif ($roll <= 74) {
	    $self->Complexion("Medium");
	} else {
	    $self->Complexion("Dark");
	}
    } elsif ($self->Species() eq 'Sindarin') {
	if ($roll <= 97) {
	    $self->Complexion("Fair");
	} elsif ($roll <= 99) {
	    $self->Complexion("Medium");
	} else {
	    $self->Complexion("Dark");
	}
    } elsif ($self->Species() eq 'Khuzdul') {
	if ($roll <= 90) {
	    $self->Complexion("Fair");
	} elsif ($roll <= 99) {
	    $self->Complexion("Medium");
	} else {
	    $self->Complexion("Dark");
	}
    } else {
	die;
    }
}
######################################################################
sub rollHairColor {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($self->Complexion() eq 'Fair') {
	$roll += 25;
    } elsif ($self->Complexion() eq 'Dark') {
	$roll -= 25;
    } elsif ($self->Complexion() ne 'Medium') {
	die "Invalid Complexion";
    }
    if ($self->Species() eq 'Human') {
	if ($roll <= 40) {
	    $self->HairColor("Brown");
	} elsif ($roll <= 55) {
	    $self->HairColor("Black");
	} elsif ($roll <= 65) {
	    $self->HairColor("Red");
	} elsif ($roll <= 70) {
	    $self->HairColor("Silver");
	} else {
	    $self->HairColor("Blonde");
	}
    } elsif ($self->Species() eq 'Sindarin') {
	if ($roll <= 25) {
	    $self->HairColor("Brown");
	} elsif ($roll <= 48) {
	    $self->HairColor("Black");
	} elsif ($roll <= 50) {
	    $self->HairColor("Red");
	} elsif ($roll <= 60) {
	    $self->HairColor("Silver");
	} else {
	    $self->HairColor("Blonde");
	}
    } elsif ($self->Species() eq 'Khuzdul') {
	if ($roll <= 40) {
	    $self->HairColor("Brown");
	} elsif ($roll <= 45) {
	    $self->HairColor("Black");
	} elsif ($roll <= 70) {
	    $self->HairColor("Red");
	} elsif ($roll <= 90) {
	    $self->HairColor("Silver");
	} else {
	    $self->HairColor("Blonde");
	}
    } else {
	die;
    }
    return $self;
}
######################################################################
sub rollEyeColor {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($self->Complexion() eq 'Fair') {
	$roll += 25;
    } elsif ($self->Complexion() eq 'Dark') {
	$roll -= 25;
    } elsif ($self->Complexion() ne 'Medium') {
	die "Invalid Complexion";
    }
    if ($self->Species() eq 'Human') {
	if ($roll <= 40) {
	    $self->EyeColor("Hazel");
	} elsif ($roll <= 55) {
	    $self->EyeColor("Gray");
	} elsif ($roll <= 56) {
	    $self->EyeColor("Violet");
	} elsif ($roll <= 70) {
	    $self->EyeColor("Green");
	} else {
	    $self->EyeColor("Blue");
	}
    } elsif ($self->Species() eq 'Sindarin') {
	if ($roll <= 15) {
	    $self->EyeColor("Hazel");
	} elsif ($roll <= 65) {
	    $self->EyeColor("Gray");
	} elsif ($roll <= 75) {
	    $self->EyeColor("Violet");
	} elsif ($roll <= 80) {
	    $self->EyeColor("Green");
	} else {
	    $self->EyeColor("Blue");
	}
    } elsif ($self->Species() eq 'Khuzdul') {
	if ($roll <= 50) {
	    $self->EyeColor("Hazel");
	} elsif ($roll <= 69) {
	    $self->EyeColor("Gray");
	} elsif ($roll <= 70) {
	    $self->EyeColor("Violet");
	} elsif ($roll <= 85) {
	    $self->EyeColor("Green");
	} else {
	    $self->EyeColor("Blue");
	}
    } else {
	die;
    }
    return $self;
}
######################################################################
sub rollStrength {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 1 if ($self->Species() eq 'Sindarin');
    $roll += 4 if ($self->Species() eq 'Khuzdul');
    my $wt = $self->Weight();
    if ($wt <= 85) {
	$roll -= 4;
    } elsif ($wt <= 110) {
	$roll -= 3;
    } elsif ($wt <= 130) {
	$roll -= 2;
    } elsif ($wt <= 145) {
	$roll -= 1;
    } elsif ($wt <= 155) {
	$roll += 0;
    } elsif ($wt <= 170) {
	$roll += 1;
    } elsif ($wt <= 190) {
	$roll += 2;
    } elsif ($wt <= 215) {
	$roll += 3;
    } elsif ($wt <= 245) {
	$roll += 4;
    } else {
	$roll += 5;
    }
    $self->Strength($roll);
    return $self;
}
######################################################################
sub rollStamina {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 1 if ($self->Species() eq 'Sindarin');
    $roll += 2 if ($self->Species() eq 'Khuzdul');
    $self->Stamina($roll);
    return $self;
}
######################################################################
sub rollDexterity {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    $roll += 1 if ($self->Species() eq 'Khuzdul');
    $self->Dexterity($roll);
    return $self;
}
######################################################################
sub rollAgility {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    if ($self->Frame() eq 'Scant') {
	$roll += 2;
    } elsif ($self->Frame() eq 'Light') {
	$roll += 1;
    } elsif ($self->Frame() eq 'Heavy') {
	$roll -= 1;
    } elsif ($self->Frame() eq 'Massive') {
	$roll -= 2;
    }
    $self->Agility($roll);
    return $self;
}
######################################################################
sub rollEyesight {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    $roll += 1 if ($self->Species() eq 'Khuzdul');
    $roll += 1 if ($self->Culture() eq 'Tribal');
    $self->Eyesight($roll);
    return $self;
}
######################################################################
sub rollHearing {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    $roll += 2 if ($self->Species() eq 'Khuzdul');
    $roll += 2 if ($self->Culture() eq 'Tribal');
    $self->Hearing($roll);
    return $self;
}
######################################################################
sub rollSmell {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 3 if ($self->Species() eq 'Sindarin');
    $roll += 2 if ($self->Species() eq 'Khuzdul');
    $roll += 2 if ($self->Culture() eq 'Tribal');
    $self->Smell($roll);
    return $self;
}
######################################################################
sub rollVoice {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 2 if ($self->Species() eq 'Sindarin');
    $self->Voice($roll);
    return $self;
}
######################################################################
sub rollIntelligence {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $self->Intelligence($roll);
    return $self;
}
######################################################################
sub rollAura {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 2 if (($self->Species() eq 'Human') and ($self->Sex() eq 'Female'));
    $roll += 4 if ($self->Species() eq 'Sindarin');
    $roll -= 2 if ($self->Species() eq 'Khuzdul');
    $self->Aura($roll);
    return $self;
}
######################################################################
sub rollWill {
    my $self = shift(@_);
    my $roll = $self->roll("best3/4d6");
    $roll += 3 if ($self->Species() eq 'Khuzdul');
    $self->Will($roll);
    return $self;
}
######################################################################
sub rollMorality() {
    my $self = shift(@_);
    my $roll = $self->roll("3d6");
    $roll += 3 if ($self->Species() eq 'Sindarin');
    $self->Morality($roll);
    return $self;
}
######################################################################
sub rollSexuality() {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($roll <= 90) { $self->Sexuality("Heterosexual"); }
    elsif ($roll <= 96) { $self->Sexuality("Bi-Sexual"); }
    else { $self->Sexuality("Homosexual (exclusively)"); }
    return $self;
}
######################################################################
sub rollDeity {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($roll <= 9) { $self->Deity("Agrik"); }
    elsif ($roll <= 18) { $self->Deity("Halea"); }
    elsif ($roll <= 27) { $self->Deity("Ilvir"); }
    elsif ($roll <= 36) { $self->Deity("Larani"); }
    elsif ($roll <= 45) { $self->Deity("Morgath"); }
    elsif ($roll <= 54) { $self->Deity("Naveh"); }
    elsif ($roll <= 63) { $self->Deity("Peoni"); }
    elsif ($roll <= 72) { $self->Deity("Sarajin"); }
    elsif ($roll <= 81) { $self->Deity("Save-K'nor"); }
    elsif ($roll <= 90) { $self->Deity("Siem"); }
    elsif ($roll <= 99) { $self->Deity("Tribal"); }
    else { $self->Deity("Godless"); }
    return $self;
}

######################################################################
sub rollPiety() {
    my $self = shift(@_);
    my $roll = $self->roll("5d6");
    $self->Piety($roll);
    return $self;
}
######################################################################
sub rollOccupation {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    my $occname = $data->getOccupationResult($self->Culture(),
					     $self->SocialClass(),
					     $roll);
    if ($occname =~ m/(.*)\|(.*)/) {
	my ($occmale,$occfemale) = ($1,$2);
	if ($self->Sex() eq 'Male') { 
	    $self->Occupation($occmale);
	} elsif ($self->Sex() eq 'Female') { 
	    $self->Occupation($occfemale);
	} else {
	    die "Invalid character sex ".$self->Sex();
	}
    } else {
	$self->Occupation($occname);
    }
    return $self;
}
######################################################################
sub getRandomTalent() {
    my $self = shift(@_);
    my $roll = $self->roll("1d100");
    if ($roll <= 1) {
	return "Charm";
    } elsif ($roll <= 9) {
	return "Clairvoyance";
    } elsif ($roll <= 14) {
	return "Disembodiment";
    } elsif ($roll <= 25) {
	return "Healing";
    } elsif ($roll <= 29) {
	return "Hex";
    } elsif ($roll <= 37) {
	return "Medium";
    } elsif ($roll <= 45) {
	return "Mental Bolt";
    } elsif ($roll <= 54) {
	return "Negation";
    } elsif ($roll <= 62) {
	return "Prescience";
    } elsif ($roll <= 71) {
	return "Psychometry";
    } elsif ($roll <= 77) {
	return "Pyrokinesis";
    } elsif ($roll <= 86) {
	return "Sensitivity";
    } elsif ($roll <= 90) {
	return "Telekinesis";
    } elsif ($roll <= 97) {
	return "Telepathy";
    } elsif ($roll == 98) {
	return "Transference";
    } elsif ($roll == 99) {
	return "Transmutation";
    } else {
	return "Unique Talent";
    }
}
######################################################################
sub roll {
    my $self = shift(@_);
    my $arg = shift(@_);
    if ($arg =~ m/^([1-9]\d*)d([1-9]\d*)$/) {
        my ($number,$sides) = ($1,$2);
        my $total = 0;
        for (1 .. $number) {
            $total += (int(rand($sides))+1);
        }
        return $total;
    } elsif ($arg =~ m/^best([1-9]\d*)\/([1-9]\d*)d([1-9]\d*)$/) {
        my ($keep,$number,$sides) = ($1,$2,$3);
        my @dice = ();
        for (1 .. $number) {
            push(@dice,(int(rand($sides))+1));
        }
        my @sorted = sort { $b <=> $a } @dice;
        my $total = 0;
        for (0 .. ($keep-1)) {
            $total += $sorted[$_];
        }
        return $total;
    } else {
        die "Illegal argument to roll '$arg'";
    }
}
######################################################################
sub validate {
    my ($self,$printflag) = @_;

    # change the internal flag for warning if a value is changed
    $self->{"_HarnChar:printIfChange"} = $printflag;

    $self->{'Endurance'} = int( ( $self->Strength() + $self->Stamina() + 
				  $self->Will() + 1) / 3 );
    $self->Sunsign( $self->getSunsign($self->BirthDate()) );
    # foreach my $statname ($self->stats()) {
    # if ($statname =~ m/Skill_(\w)/) {
    # }
    # }

    # change the internal flag for warning if a value is changed
    $self->{"_HarnChar:printIfChange"} = 0;

    return 1;
} # end of sub validate

######################################################################
sub stringify {
    my $self = shift(@_);
    my $text;
    $text .= $self->Species() . " " . $self->Sex() . ", born " . 
	$self->BirthDate() . " (Sunsign " . $self->Sunsign() . ")\n";
    $text .= "Social: " . $self->Culture() . " " . $self->SocialClass() . "\n";
    $text .= "Family: " . $self->SiblingRank() . ", " . $self->Parentage() . "\n";
    $text .= "Estrangement: " . $self->Estrangement() . "\n";
    $text .= "Clanhead: " . $self->Clanhead() . "\n";
    $text .= "Height: " . $self->Height() . ", ";
    $text .= "Frame: " . $self->Frame() . ", ";
    $text .= "Weight: " . $self->Weight() . "\n";
    $text .= "Comliness: " . $self->Comliness() . "\n";
    $text .= "Complexion: " . $self->Complexion() . ", " . 
	"Hair Color: " . $self->HairColor() . ", " . 
	"Eye Color: " . $self->EyeColor() . "\n";
    $text .= "Strength: " . $self->Strength() . "\n";
    $text .= "Stamina: " . $self->Stamina() . "\n";
    $text .= "Dexterity: " . $self->Dexterity() . "\n";
    $text .= "Agility: " . $self->Agility() . "\n";
    $text .= "Eyesight: " . $self->Eyesight() . "\n";
    $text .= "Hearing: " . $self->Hearing() . "\n";
    $text .= "Smell: " . $self->Smell() . "\n";
    $text .= "Voice: " . $self->Voice() . "\n";
    $text .= "Intelligence: " . $self->Intelligence() . "\n";
    $text .= "Aura: " . $self->Aura() . "\n";
    $text .= "Will: " . $self->Will() . "\n";
    $text .= "Morality: " . $self->Morality() . "\n";
    $text .= "Piety: " . $self->Piety() . "\n";
    $text .= "Occupation: " . $self->Occupation() . "\n";
    $text .= "Skills:\n";
    foreach my $type (sort $data->getSkillTypeList()) {
	my $prefix = "Skill_" . $type . "_";
	for (my $n = 1; defined($self->{$prefix.$n."_Name"}); $n++) {
	    $text .= sprintf(" - %-25s (SB %2d) %2d\n",
			     $self->{$prefix.$n."_Name"},
			     $self->{$prefix.$n."_SB"},
			     $self->{$prefix.$n."_ML"});
	}
    }
    return $text;
}
######################################################################

sub _statcmp {
    my ($arga,$argb) = @_;
    my $max = scalar(@stats);
    my ($orda,$ordb) = (_statord($arga),_statord($argb));
    if (($orda < $max) or ($ordb < $max)) {
	return ($orda <=> $ordb);
    } else {
	# they are both skills or freeform stats
	return ($arga cmp $argb);
    }
}
sub _statord {
    my $arg = shift;
    for (0..$#stats) { 
	return $_ if ($arg eq $stats[$_]); 
    }
    return $#stats + 2;
}

######################################################################

1; # required end-of-module line

